#include <hikari/output_power.h>

#include <assert.h>

#include <wlr/types/wlr_seat.h>
#include <wlr/util/log.h>

#include <hikari/action.h>
#include <hikari/node.h>
#include <hikari/normal_mode.h>
#include <hikari/output.h>
// #include <hikari/renderer.h>
#include <hikari/server.h>

static struct hikari_output_power_state *
get_state(void)
{
    struct hikari_output_power_state *state = &hikari_server.output_power_state;
    return state;
}

static void
enable_outputs(void)
{
    wlr_log(WLR_DEBUG, "hikari %s", __func__);
    struct hikari_output_power_state *state = get_state();
    wlr_log(WLR_DEBUG, "hikari %s: step 1", __func__);
    if (!state->outputs_disabled) {
        return;
    }

    struct hikari_output *output;
    wl_list_for_each (output, &hikari_server.outputs, server_outputs) {
        wlr_log(WLR_DEBUG, "hikari %s: step 2", __func__);
        struct wlr_output_state pending = {0};
        wlr_output_state_set_enabled(&pending, true);
        wlr_output_commit_state(output->wlr_output, &pending);
        wlr_output_state_finish(&pending);
        output->enabled = true;
        hikari_output_damage_whole(output);
    }
    state->outputs_disabled = false;
    wlr_log(WLR_DEBUG, "hikari %s: step 3", __func__);
    hikari_output_power_state_fini();
    hikari_cursor_reset_image(&hikari_server.cursor);
}

static void
keyboard_grab_enter(
        struct wlr_seat_keyboard_grab *grab,
        struct wlr_surface *surface,
        const uint32_t keycodes[],
        size_t num_keycodes,
        const struct wlr_keyboard_modifiers *modifiers
        )
{
}

static void
keyboard_grab_clear_focus(struct wlr_seat_keyboard_grab *grab)
{
}

static void
keyboard_grab_key(struct wlr_seat_keyboard_grab *grab,
        uint32_t time_msec, uint32_t key, uint32_t state)
{
    struct hikari_output_power_state *power_state = get_state();

    if (power_state->key_init) {
        if (state == WL_KEYBOARD_KEY_STATE_RELEASED) {
            power_state->key_init = false;
        }
        return;
    }
    if (power_state->outputs_disabled) {
        enable_outputs();
        hikari_output_power_state_fini();
    }
}

static void
keyboard_grab_modifiers(struct wlr_seat_keyboard_grab *grab,
        const struct wlr_keyboard_modifiers *modifiers)
{
    enable_outputs();
    hikari_output_power_state_fini();
}

static void
keyboard_grab_cancel(struct wlr_seat_keyboard_grab *grab)
{
}

static void
power_cursor_enter(struct wlr_seat_pointer_grab *grab,
        struct wlr_surface *surface, double sx, double sy)
{
}

static void
power_cursor_clear_focus(struct wlr_seat_pointer_grab *grab)
{
}

static void
power_cursor_motion(struct wlr_seat_pointer_grab *grab,
            uint32_t time_msec, double sx, double sy)
{
    enable_outputs();
    hikari_output_power_state_fini();
}

static uint32_t
power_cursor_button(struct wlr_seat_pointer_grab *grab, uint32_t time_msec,
        uint32_t button, enum wlr_button_state state)
{
    enable_outputs();
    hikari_output_power_state_fini();
    return 0;
}

static void
power_cursor_axis(struct wlr_seat_pointer_grab *grab, uint32_t time_msec,
        enum wlr_axis_orientation orientation, double value,
        int32_t value_discrete, enum wlr_axis_source source)
{
}

static void
power_cursor_frame(struct wlr_seat_pointer_grab *grab)
{
}

static void
power_cursor_cancel(struct wlr_seat_pointer_grab *grab)
{
}

void
hikari_output_power_manager_set_mode_handler(
        struct wl_listener *listener, void *data)
{
    wlr_log(WLR_DEBUG, "hikari %s", __func__);
    struct wlr_output_power_v1_set_mode_event *event = data;
    struct wlr_output *wlr_output = event->output;
    struct hikari_output *output = NULL;
    struct hikari_output_power_state *state = get_state();
    struct wlr_output_state pending = {0};
    struct hikari_output *each_output;

    wl_list_for_each (each_output, &hikari_server.outputs, server_outputs) {
        if (wlr_output == each_output->wlr_output) {
            output = each_output;
        }
    }
    assert(wlr_output == output->wlr_output);

    wlr_seat_pointer_notify_clear_focus(hikari_server.seat);
    wlr_seat_keyboard_notify_clear_focus(hikari_server.seat);
    if (event->mode == ZWLR_OUTPUT_POWER_V1_MODE_ON) {
        if (state->outputs_disabled) {
            enable_outputs();
            hikari_output_power_state_fini();
            state->outputs_disabled = false;
        }
        else {
            wlr_log(WLR_DEBUG, "hikari %s: turns on %s",
                    __func__, wlr_output->name);
            wlr_output_state_set_enabled(&pending, true);
            wlr_output_commit_state(wlr_output, &pending);
            wlr_output_state_finish(&pending);
            output->enabled = true;
            hikari_cursor_reset_image(&hikari_server.cursor);
        }
    }
    else if (event->mode == ZWLR_OUTPUT_POWER_V1_MODE_OFF) {
        wlr_log(WLR_DEBUG, "hikari %s: turns off %s",
                __func__, wlr_output->name);

        wlr_output_state_set_enabled(&pending, false);
        wlr_output_commit_state(wlr_output, &pending);
        wlr_output_state_finish(&pending);
        output->enabled = false;

        bool outputs_enabled = false;
        wl_list_for_each (each_output, &hikari_server.outputs, server_outputs) {
            wlr_log(WLR_DEBUG, "hikari %s: check %s",
                    __func__, each_output->wlr_output->name);
            outputs_enabled |= each_output->enabled;
        }
        if (!outputs_enabled) {
            hikari_output_power_state_enter();
            state->outputs_disabled = true;
        }
    }
}

void
hikari_output_power_state_init(struct hikari_output_power_state *output_power_state)
{
    output_power_state->outputs_disabled = false;
    output_power_state->key_init = false;

    output_power_state->keyboard_grab_interface.enter = keyboard_grab_enter;
    output_power_state->keyboard_grab_interface.clear_focus
        = keyboard_grab_clear_focus;
    output_power_state->keyboard_grab_interface.key = keyboard_grab_key;
    output_power_state->keyboard_grab_interface.modifiers
        = keyboard_grab_modifiers;
    output_power_state->keyboard_grab_interface.cancel = keyboard_grab_cancel;
    output_power_state->keyboard_grab.interface
        = &output_power_state->keyboard_grab_interface;
    output_power_state->keyboard_grab.seat = hikari_server.seat;
    output_power_state->keyboard_grab.data = NULL;

    output_power_state->cursor_grab_interface.enter = power_cursor_enter;
    output_power_state->cursor_grab_interface.clear_focus
        = power_cursor_clear_focus;
    output_power_state->cursor_grab_interface.motion = power_cursor_motion;
    output_power_state->cursor_grab_interface.button = power_cursor_button;
    output_power_state->cursor_grab_interface.axis = power_cursor_axis;
    output_power_state->cursor_grab_interface.frame = power_cursor_frame;
    output_power_state->cursor_grab_interface.cancel = power_cursor_cancel;
    output_power_state->cursor_grab.interface
        = &output_power_state->cursor_grab_interface;
    output_power_state->cursor_grab.seat = hikari_server.seat;
    output_power_state->cursor_grab.data = NULL;
}

void
hikari_output_power_state_enter(void)
{
    struct hikari_output_power_state *state = get_state();
    struct wlr_seat *seat = hikari_server.seat;

    state->key_init = true;
    wlr_seat_pointer_start_grab(seat, &state->cursor_grab);
    wlr_seat_keyboard_start_grab(seat, &state->keyboard_grab);
}

void
hikari_output_power_state_fini(void)
{
    struct hikari_server *server = &hikari_server;
    struct hikari_cursor *cursor = &server->cursor;
    struct wlr_seat *seat = server->seat;
    struct hikari_node *node = NULL;
    struct hikari_workspace *workspace = NULL;
    struct wlr_surface *surface = NULL;
    double sx, sy;

    wlr_seat_pointer_end_grab(seat);
    wlr_seat_keyboard_end_grab(seat);
    hikari_server_cursor_focus();
    node = hikari_server_node_at(
            cursor->wlr_cursor->x, cursor->wlr_cursor->y,
            &surface, &workspace, &sx, &sy);
    if (node != NULL) {
        wlr_seat_pointer_notify_enter(seat, surface, sx, sy);
        wlr_seat_pointer_notify_motion(seat, 1, sx, sy);
        wlr_seat_keyboard_notify_enter(seat, surface, 0, 0, NULL);
    }
}
